from os.path import join, basename
from random import random, choice, randint
from glob import glob
from itertools import chain

from pygame.mixer import Sound
from pygame.image import load
from pygame.transform import rotate, flip, scale2x
from pygame.event import clear
from pygame.locals import *

from lib.pgfw.pgfw.Game import Game
from lib.pgfw.pgfw.GameChild import GameChild
from lib.pgfw.pgfw.Animation import Animation
from lib.pgfw.pgfw.Sprite import Sprite
from lib.pgfw.pgfw.extension import get_hue_shifted_surface, get_inverted_surface

class SoundEffects(GameChild):

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        effects = self.effects = []
        for path in glob(join(self.get_resource("aud/effect"), "*.wav")):
            effects.append(SoundEffect(self, path))

    def play(self, name, volume=None, x=None):
        for effect in self.effects:
            if effect.name == name:
                effect.play(volume=volume, x=x)


class SoundEffect(GameChild, Sound):

    DEFAULT_VOLUME = .5

    def __init__(self, parent, path, volume=DEFAULT_VOLUME):
        GameChild.__init__(self, parent)
        Sound.__init__(self, path)
        self.name = basename(path).split(".")[0]
        self.display_surface = self.get_display_surface()
        self.set_volume(volume)

    def play(self, loops=0, maxtime=0, fade_ms=0, position=None, x=None, volume=None):
        channel = Sound.play(self, loops, maxtime, fade_ms)
        if x is not None:
            position = float(x) / self.display_surface.get_width()
	if position is not None and channel is not None:
            channel.set_volume(*self.get_panning(position))
        return channel

    def get_panning(self, position):
        return 1 - max(0, ((position - .5) * 2)), \
               1 + min(0, ((position - .5) * 2))


class DirtbagGame(Game):

    AMBIENT_AUDIO_CHANCE = .004

    def __init__(self):
        Game.__init__(self)
        self.suppressing_input = False
        self.sound_effects = SoundEffects(self)
        self.subscribe(self.respond)
        self.subscribe(self.respond, MOUSEBUTTONDOWN)
        self.background = load(join("resource", "img", "background.png")).convert()
        self.ouroboros = Ouroboros(self)
        self.counter = Counter(self)
        self.hands = Hands(self)
        clear()

    def respond(self, event):
        if not self.suppressing_input:
            if self.delegate.compare(event, "any") or (event.type == MOUSEBUTTONDOWN and
                                                       event.button in (1, 3)):
                self.suppressing_input = True
                self.hands.give()
                self.sound_effects.play("interact-3")

    def update(self):
        ds = self.get_display_surface()
        if random() < self.AMBIENT_AUDIO_CHANCE:
            self.sound_effects.play(choice(("clear-0", "clear-1", "scratch")),
                                    x=randint(0, ds.get_width()))
        ds.blit(self.background, (0, 0))
        self.hands.update()
        self.ouroboros.update()
        self.counter.update()


class Ouroboros(Animation):

    HEAD_INITIAL_OFFSET = 2
    SLITHER_INTERVAL = 750
    N, W, S, E = range(4)
    MARGIN = 4
    BLINK_INTERVAL = 200
    SWALLOW_DELAY = 3100
    MIRROR_INTERVAL = 200
    MIRROR_LENGTH = 3000

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.load_pieces()
        ds = self.display_surface = self.get_display_surface()
        thickness = self.clip_thickness = self.segment[0][0][0].get_height()
        self.horizontal_length = ds.get_width() - self.MARGIN * 3 - thickness
        self.vertical_length = ds.get_height() - self.MARGIN * 3 - thickness
        gap_length = self.gap_length = self.segment[0][0][0].get_width() * 2
        self.available_gap_positions = self.horizontal_length / gap_length * 2 + \
                                       self.vertical_length / gap_length * 2
        self.slithered = False
        self.tail_visible = True
        self.mirrored = False
        self.head_offset = self.HEAD_INITIAL_OFFSET
        self.register(self.slither, interval=self.SLITHER_INTERVAL)
        self.register(self.blink_tail, interval=self.BLINK_INTERVAL)
        self.register(self.swallow, self.end_mirror)
        self.register(self.mirror, interval=self.MIRROR_INTERVAL)
        self.play(self.slither)

    def load_pieces(self):
        segment, head, tail = self.segment, self.head, self.tail = [], [], []
        base_segment = load(join("resource", "img", "ouroboros-segment.png")).convert_alpha()
        base_head = load(join("resource", "img", "ouroboros-head.png")).convert_alpha()
        base_head_raised = load(join("resource", "img", "ouroboros-head-raised.png")).convert_alpha()
        base_tail = load(join("resource", "img", "ouroboros-tail.png")).convert_alpha()
        for side in self.N, self.W, self.S, self.E:
            angle = 90 * side
            segment.append([[], []])
            head.append([[], []])
            tail.append([[], []])
            rotated_base_segment = rotate(base_segment, angle)
            rotated_base_head = rotate(base_head, angle)
            rotated_base_tail = rotate(base_tail, angle)
            segment[side][0].append(rotated_base_segment)
            segment[side][0].append(get_inverted_surface(rotated_base_segment))
            head[side][0].append(rotated_base_head)
            head[side][0].append(get_inverted_surface(rotated_base_head))
            tail[side][0].append(rotated_base_tail)
            tail[side][0].append(get_inverted_surface(rotated_base_tail))
            x, y = side in (self.W, self.E), side in (self.N, self.S)
            segment[side][1].append(flip(rotated_base_segment, x, y))
            segment[side][1].append(get_inverted_surface(flip(rotated_base_segment, x, y)))
            head[side][1].append(rotate(base_head_raised, angle))
            head[side][1].append(get_inverted_surface(rotate(base_head_raised, angle)))
            tail[side][1].append(flip(rotated_base_tail, x, y))
            tail[side][1].append(get_inverted_surface(flip(rotated_base_tail, x, y)))

    def slither(self):
        self.slithered = not self.slithered

    def eat(self):
        self.play(self.blink_tail)
        self.play(self.swallow, delay=self.SWALLOW_DELAY, play_once=True)

    def blink_tail(self):
        self.tail_visible = not self.tail_visible

    def swallow(self):
        self.head_offset += 1
        self.tail_visible = True
        self.halt(self.blink_tail)
        self.get_game().sound_effects.play("eat")

    def mirror(self):
        self.mirrored = not self.mirrored

    def end_mirror(self):
        self.mirrored = False
        self.halt(self.mirror)
        self.eat()

    def update(self):
        Animation.update(self)
        ds = self.display_surface
        vl, hl, thickness = self.vertical_length, self.horizontal_length, self.clip_thickness
        sl = self.segment[0][0][0].get_width()
        position = 0
        head_offset = self.head_offset % self.available_gap_positions
        for side in self.N, self.W, self.S, self.E:
            segment_surface = self.segment[side][self.slithered][self.mirrored]
            head_surface = self.head[side][self.slithered][self.mirrored]
            tail_surface = self.tail[side][self.slithered][self.mirrored]
            if side == self.N:
                clip = Rect(self.MARGIN, self.MARGIN, hl, thickness)
                ds.set_clip(clip)
                y = self.MARGIN
                remainder = clip.w % self.gap_length
                x = clip.right - remainder
                first = True
                while x >= clip.left:
                    if first or position != head_offset:
                        ds.blit(segment_surface, (x, y))
                        ds.blit(segment_surface, (x + sl, y))
                    elif position == head_offset:
                        ds.blit(head_surface, (x + sl, y))
                        if self.tail_visible:
                            ds.blit(tail_surface, (x, y))
                    if not first:
                        position += 1
                    first = False
                    x -= self.gap_length
            elif side == self.W:
                clip = Rect(self.MARGIN, thickness + self.MARGIN * 2, thickness, vl)
                ds.set_clip(clip)
                x = self.MARGIN
                remainder = clip.h % self.gap_length
                y = clip.top - (self.gap_length - remainder)
                first = True
                while y <= clip.bottom - self.gap_length:
                    if first or position != head_offset:
                        ds.blit(segment_surface, (x, y))
                        ds.blit(segment_surface, (x, y + sl))
                    elif position == head_offset:
                        ds.blit(head_surface, (x, y))
                        if self.tail_visible:
                            ds.blit(tail_surface, (x, y + sl))
                    if not first:
                        position += 1
                    first = False
                    y += self.gap_length
            elif side == self.S:
                clip = Rect(thickness + self.MARGIN * 2,
                            ds.get_height() - thickness - self.MARGIN, hl, thickness)
                ds.set_clip(clip)
                y = clip.top
                remainder = clip.w % self.gap_length
                x = clip.left - (self.gap_length - remainder)
                first = True
                while x <= clip.right - self.gap_length:
                    if first or position != head_offset:
                        ds.blit(segment_surface, (x, y))
                        ds.blit(segment_surface, (x + sl, y))
                    elif position == head_offset:
                        ds.blit(head_surface, (x, y))
                        if self.tail_visible:
                            ds.blit(tail_surface, (x + sl, y))
                    if not first:
                        position += 1
                    first = False
                    x += self.gap_length
            elif side == self.E:
                clip = Rect(ds.get_width() - thickness - self.MARGIN, self.MARGIN, thickness, vl)
                ds.set_clip(clip)
                x = clip.left
                remainder = clip.h % self.gap_length
                y = clip.bottom - remainder
                first = True
                while y >= clip.top:
                    if first or position != head_offset:
                        ds.blit(segment_surface, (x, y))
                        ds.blit(segment_surface, (x, y + sl))
                    elif position == head_offset:
                        ds.blit(head_surface, (x, y + sl))
                        if self.tail_visible:
                            ds.blit(tail_surface, (x, y))
                    if not first:
                        position += 1
                    first = False
                    y -= self.gap_length
            ds.set_clip()


class Counter(Animation):

    INITIAL_COUNT = 0
    POSITION = 720, 60
    MARGIN = 4
    HUE_STEP = 10
    FLASH_LENGTH = 3000
    ENLARGE_LENGTH = 2000

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.count = self.INITIAL_COUNT
        self.enlarged = False
        self.set_sprites()
        self.register(self.end_flash, self.end_enlarge)

    def set_sprites(self):
        regular_sprites = self.regular_sprites = []
        large_sprites = self.large_sprites = []
        for path in sorted(glob(join("resource", "img", "digit-*.png"))):
            regular_sprites.append(Sprite(self))
            large_sprites.append(Sprite(self))
            base = load(path).convert_alpha()
            large = scale2x(base)
            regular_sprites[-1].add_frame(base)
            large_sprites[-1].add_frame(large)
            regular_sprites[-1].hide()
            large_sprites[-1].hide()
            for offset in xrange(self.HUE_STEP, 360, self.HUE_STEP):
                regular_sprites[-1].add_frame(get_hue_shifted_surface(base, offset))
                large_sprites[-1].add_frame(get_hue_shifted_surface(large, offset))
            regular_sprites[-1].add_frameset([0], name="static", switch=True)
            large_sprites[-1].add_frameset([0], name="static", switch=True)
        regular_sprites[0].location.topleft = self.POSITION
        large_sprites[0].location.topleft = self.POSITION[0] - regular_sprites[0].location.w, \
                                            self.POSITION[1]
        regular_sprites[0].unhide()

    def increment(self):
        self.count += 1
        for sprites in (self.regular_sprites, self.large_sprites):
            for sprite in sprites:
                sprite.set_frameset(0)
                sprite.hide()
                sprite.remove_locations()
                sprite.location.bottomright = 0, 0
        text = str(self.count)
        for sprites in (self.regular_sprites, self.large_sprites):
            if sprites == self.regular_sprites:
                x = self.POSITION[0]
            else:
                x = self.POSITION[0] - self.regular_sprites[0].location.w
            start = x - (len(text) - 1) * (self.MARGIN + sprites[0].location.w)
            ii = 0
            while start <= self.POSITION[0]:
                sprite = sprites[int(text[ii])]
                sprite.unhide()
                sprite.add_location((start, self.POSITION[1]))
                start += sprite.location.w + self.MARGIN
                ii += 1
        self.play(self.end_flash, delay=self.FLASH_LENGTH, play_once=True)
        self.enlarge()

    def end_flash(self):
        for sprite in chain(self.regular_sprites, self.large_sprites):
            sprite.set_frameset("static")

    def enlarge(self):
        self.enlarged = True
        self.play(self.end_enlarge, delay=self.ENLARGE_LENGTH, play_once=True)

    def end_enlarge(self):
        self.enlarged = False

    def update(self):
        Animation.update(self)
        for sprite in self.large_sprites if self.enlarged else self.regular_sprites:
            sprite.update()


class Hands(Animation):

    WAIT_BOTTOM_POSITION = 416, 225
    WAIT_TOP_POSITION = 184, 0
    WAIT_SHIFT_CHANCE = .006
    WAIT_SHIFT_SPEED = .3
    WAIT_SHIFT_DEPTH = 3
    GIVE_BOTTOM_POSITION = 399, 216
    GIVE_TOP_POSITION = 245, 0
    GIVE_RETRIEVE_SPEED = 4
    GIVE_RETURN_SPEED = 4
    GIVE_FLIP_CUE = 280
    GIVE_RETURN_DELAY = 800
    CARD_RETURN_OFFSET = 14, 202
    CARD_TAKE_OFFSET = -54, 76
    TRANSFER_DELAY = 3000
    STORE_DELAY = 2000
    STORE_SPEED = 4
    REMARK_POSITION = 88, 402
    REMARK_LENGTH = 3000
    RECENTER_SPEED = 5
    WALK_SPEED = 8
    WALK_DELAY = 800

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.waiting = True
        self.set_sprites()
        self.register(self.wait_shift_bottom, self.wait_shift_top,
                      self.retrieve, self.bring, self.transfer, self.store,
                      self.end_remark, self.recenter, self.walk)

    def set_sprites(self):
        wait = load(join("resource", "img", "hand-wait.png")).convert_alpha()
        wait_bottom = self.wait_bottom = Sprite(self)
        wait_bottom.add_frame(wait)
        wait_top = self.wait_top = Sprite(self)
        wait_top.add_frame(rotate(wait, 180))
        self.reset_wait_positions()
        give = load(join("resource", "img", "hand-give.png")).convert_alpha()
        give_bottom = self.give_bottom = Sprite(self)
        give_bottom.add_frame(give)
        give_bottom.location.topleft = self.GIVE_BOTTOM_POSITION
        give_bottom.hide()
        give_top = self.give_top = Sprite(self)
        give_top.add_frame(rotate(give, 180))
        give_top.location.bottomleft = self.GIVE_TOP_POSITION
        give_top.hide()
        card = self.card = Sprite(self)
        card.add_frame(load(join("resource", "img", "card-2.png")).convert_alpha())
        card.hide()
        remark = self.remark = Sprite(self)
        remark.add_frame(load(join("resource", "img", "remark.png")).convert_alpha())
        remark.location.topleft = self.REMARK_POSITION
        remark.hide()

    def reset_wait_positions(self):
        self.wait_bottom.location.topleft = self.WAIT_BOTTOM_POSITION
        self.wait_top.location.topleft = self.WAIT_TOP_POSITION
        self.wait_shift_top_retreating = True
        self.wait_shift_bottom_retreating = True

    def wait_shift_top(self):
        if self.wait_shift_top_retreating:
            if self.wait_top.location.top <= -self.WAIT_SHIFT_DEPTH:
                self.wait_shift_top_retreating = False
            else:
                self.wait_top.move(dy=-self.WAIT_SHIFT_SPEED)
        else:
            if self.wait_top.location.top >= self.WAIT_TOP_POSITION[1]:
                self.halt(self.wait_shift_top)
                self.wait_top.location.topleft = self.WAIT_TOP_POSITION
                self.wait_shift_top_retreating = True
            else:
                self.wait_top.move(dy=self.WAIT_SHIFT_SPEED)

    def wait_shift_bottom(self):
        if self.wait_shift_bottom_retreating:
            if self.wait_bottom.location.bottom > self.get_display_surface().get_height() + \
               self.WAIT_SHIFT_DEPTH + 2:
                self.wait_shift_bottom_retreating = False
            else:
                self.wait_bottom.move(dy=self.WAIT_SHIFT_SPEED)
        else:
            if self.wait_bottom.location.top <= self.WAIT_BOTTOM_POSITION[1]:
                self.halt(self.wait_shift_bottom)
                self.wait_bottom.location.topleft = self.WAIT_BOTTOM_POSITION
                self.wait_shift_bottom_retreating = True
            else:
                self.wait_bottom.move(dy=-self.WAIT_SHIFT_SPEED)

    def give(self):
        self.waiting = False
        self.give_bottom.location.topleft = self.GIVE_BOTTOM_POSITION
        self.halt(self.wait_shift_top)
        self.halt(self.wait_shift_bottom)
        self.reset_wait_positions()
        self.play(self.retrieve)

    def retrieve(self):
        if self.wait_top.location.bottom <= 0:
            self.wait_top.hide()
            self.give_top.location.bottom = 0
            self.give_top.unhide()
            self.card.unhide()
            self.card.location.bottom = 0
            self.halt(self.retrieve)
            self.play(self.bring, delay=self.GIVE_RETURN_DELAY)
            self.get_game().sound_effects.play("retrieve")
        else:
            self.wait_top.move(dy=-self.GIVE_RETRIEVE_SPEED)

    def bring(self):
        if self.give_top.location.top >= self.GIVE_TOP_POSITION[1]:
            self.give_top.location.top = self.GIVE_TOP_POSITION[1]
            self.halt(self.bring)
            self.play(self.transfer, delay=self.TRANSFER_DELAY, play_once=True)
        else:
            if self.give_bottom.is_hidden() and self.give_top.location.bottom >= self.GIVE_FLIP_CUE:
                self.give_bottom.unhide()
                self.wait_bottom.hide()
            self.give_top.move(dy=self.GIVE_RETURN_SPEED)
        self.card.location.topleft = self.give_top.location.move(*self.CARD_RETURN_OFFSET).topleft

    def transfer(self):
        self.card.location.topleft = self.give_bottom.location.move(*self.CARD_TAKE_OFFSET).topleft
        self.remark.hide()
        self.get_game().counter.increment()
        self.get_game().sound_effects.play("increment")
        self.remark.unhide()
        self.get_game().sound_effects.play("remark")
        self.play(self.store, delay=self.STORE_DELAY)
        self.play(self.end_remark, delay=self.REMARK_LENGTH, play_once=True)

    def store(self):
        if self.give_bottom.location.top >= self.get_display_surface().get_height():
            self.give_top.hide()
            self.wait_top.unhide()
            self.wait_top.location.topleft = self.WAIT_TOP_POSITION
            self.give_bottom.hide()
            self.wait_bottom.unhide()
            self.wait_bottom.location.top = self.get_display_surface().get_height()
            self.card.hide()
            self.halt(self.store)
            self.play(self.recenter)
            ouroboros = self.get_game().ouroboros
            ouroboros.play(ouroboros.mirror)
            ouroboros.play(ouroboros.end_mirror, delay=Ouroboros.MIRROR_LENGTH, play_once=True)
        else:
            self.give_bottom.move(dy=self.STORE_SPEED)
        self.card.location.topleft = self.give_bottom.location.move(*self.CARD_TAKE_OFFSET).topleft

    def end_remark(self):
        self.remark.hide()

    def recenter(self):
        if self.wait_bottom.location.top <= self.WAIT_BOTTOM_POSITION[1]:
            self.reset_wait_positions()
            self.halt(self.recenter)
            self.walk_looped = False
            self.play(self.walk, delay=self.WALK_DELAY)
            self.get_game().sound_effects.play("walk")
        else:
            self.wait_bottom.move(dy=-self.RECENTER_SPEED)

    def walk(self):
        ds = self.get_display_surface()
        if self.walk_looped and self.wait_bottom.location.left >= self.WAIT_BOTTOM_POSITION[0]:
            self.reset_wait_positions()
            self.get_game().suppressing_input = False
            self.waiting = True
            self.halt(self.walk)
        else:
            self.wait_bottom.move(self.WALK_SPEED)
            self.wait_top.move(-self.WALK_SPEED)
            if self.wait_bottom.location.left >= ds.get_width():
                self.walk_looped = True
                self.wait_bottom.location.left -= ds.get_width() + self.wait_bottom.location.w
            if self.wait_top.location.right <= 0:
                self.wait_top.location.right += ds.get_width() + self.wait_top.location.w

    def update(self):
        Animation.update(self)
        if self.waiting:
            if random() <= self.WAIT_SHIFT_CHANCE:
                if random() < .5:
                    if not self.is_playing(self.wait_shift_top):
                        self.play(self.wait_shift_top)
                else:
                    if not self.is_playing(self.wait_shift_bottom):
                        self.play(self.wait_shift_bottom)
        self.wait_top.update()
        self.wait_bottom.update()
        self.give_top.update()
        self.give_bottom.update()
        self.remark.update()
        self.card.update()
